@@ -27,6 +27,7 @@ DEALINGS IN THE SOFTWARE.  */
 
 #include <stdlib.h>
 #include <string.h>
+#include <strings.h>
 #include <stdio.h>
 #include <unistd.h>
 #include <math.h>
@@ -34,14 +35,23 @@ DEALINGS IN THE SOFTWARE.  */
 #include <stdbool.h>
 #include <assert.h>
 #include <getopt.h>
+#include <ctype.h>
 #include "htslib/sam.h"
 #include "htslib/faidx.h"
 #include "htslib/kstring.h"
 #include "htslib/khash.h"
+#include "htslib/klist.h"
 #include "htslib/thread_pool.h"
+#include "htslib/bgzf.h"
 #include "samtools.h"
 #include "sam_opts.h"
+
+#define DEFAULT_BARCODE_TAG "BC"
+#define DEFAULT_QUALITY_TAG "QT"
+
 KHASH_SET_INIT_STR(rg)
+#define taglist_free(p)
+KLIST_INIT(ktaglist, char*, taglist_free)
 
 typedef khash_t(rg) *rghash_t;
 
@@ -51,6 +61,7 @@ typedef struct samview_settings {
     int min_mapQ;
     int flag_on;
     int flag_off;
+    int flag_alloff;
     int min_qlen;
     int remove_B;
     uint32_t subsam_seed;
@@ -84,6 +95,8 @@ static int process_aln(const bam_hdr_t *h, bam1_t *b, samview_settings_t* settin
     }
     if (b->core.qual < settings->min_mapQ || ((b->core.flag & settings->flag_on) != settings->flag_on) || (b->core.flag & settings->flag_off))
         return 1;
+    if (settings->flag_alloff && ((b->core.flag & settings->flag_alloff) == settings->flag_alloff))
+        return 1;
     if (settings->bed && (b->core.tid < 0 || !bed_overlap(settings->bed, h->target_name[b->core.tid], b->core.pos, bam_endpos(b))))
         return 1;
     if (settings->subsam_frac > 0.) {
@@ -247,6 +260,7 @@ int main_samview(int argc, char *argv[])
         .min_mapQ = 0,
         .flag_on = 0,
         .flag_off = 0,
+        .flag_alloff = 0,
         .min_qlen = 0,
         .remove_B = 0,
         .subsam_seed = 0,
@@ -264,7 +278,7 @@ int main_samview(int argc, char *argv[])
     strcpy(out_mode, "w");
     strcpy(out_un_mode, "w");
     while ((c = getopt_long(argc, argv,
-                            "SbBcCt:h1Ho:O:q:f:F:ul:r:?T:R:L:s:@:m:x:U:",
+                            "SbBcCt:h1Ho:O:q:f:F:G:ul:r:?T:R:L:s:@:m:x:U:",
                             lopts, NULL)) >= 0) {
         switch (c) {
         case 's':
@@ -288,6 +302,7 @@ int main_samview(int argc, char *argv[])
         case 'U': fn_un_out = strdup(optarg); break;
         case 'f': settings.flag_on |= strtol(optarg, 0, 0); break;
         case 'F': settings.flag_off |= strtol(optarg, 0, 0); break;
+        case 'G': settings.flag_alloff |= strtol(optarg, 0, 0); break;
         case 'q': settings.min_mapQ = atoi(optarg); break;
         case 'u': compress_level = 0; break;
         case '1': compress_level = 1; break;
@@ -569,8 +584,9 @@ static int usage(FILE *fp, int exit_status, int is_long_help)
 "  -l STR   only include reads in library STR [null]\n"
 "  -m INT   only include reads with number of CIGAR operations consuming\n"
 "           query sequence >= INT [0]\n"
-"  -f INT   only include reads with all bits set in INT set in FLAG [0]\n"
-"  -F INT   only include reads with none of the bits set in INT set in FLAG [0]\n"
+"  -f INT   only include reads with all  of the FLAGs in INT present [0]\n"       //   F&x == x
+"  -F INT   only include reads with none of the FLAGS in INT present [0]\n"       //   F&x == 0
+"  -G INT   only EXCLUDE reads with all  of the FLAGs in INT present [0]\n"       // !(F&x == x)
 "  -s FLOAT subsample reads (given INT.FRAC option value, 0.FRAC is the\n"
 "           fraction of templates/read pairs to keep; INT part sets seed)\n"
 // read processing
@@ -649,21 +665,42 @@ static void bam2fq_usage(FILE *to, const char *command)
 "Usage: samtools %s [options...] <in.bam>\n", command);
     fprintf(to,
 "Options:\n"
-"  -0 FILE   write paired reads flagged both or neither READ1 and READ2 to FILE\n"
-"  -1 FILE   write paired reads flagged READ1 to FILE\n"
-"  -2 FILE   write paired reads flagged READ2 to FILE\n"
-"  -f INT    only include reads with all bits set in INT set in FLAG [0]\n"
-"  -F INT    only include reads with none of the bits set in INT set in FLAG [0]\n"
-"  -n        don't append /1 and /2 to the read name\n");
+"  -0 FILE              write paired reads flagged both or neither READ1 and READ2 to FILE\n"
+"  -1 FILE              write paired reads flagged READ1 to FILE\n"
+"  -2 FILE              write paired reads flagged READ2 to FILE\n"
+"  -f INT               only include reads with all  of the FLAGs in INT present [0]\n"       //   F&x == x
+"  -F INT               only include reads with none of the FLAGS in INT present [0]\n"       //   F&x == 0
+"  -G INT               only EXCLUDE reads with all  of the FLAGs in INT present [0]\n"       // !(F&x == x)
+"  -n                   don't append /1 and /2 to the read name\n"
+"  -N                   always append /1 and /2 to the read name\n");
     if (fq) fprintf(to,
-"  -O        output quality in the OQ tag if present\n");
+"  -O                   output quality in the OQ tag if present\n");
     fprintf(to,
-"  -s FILE   write singleton reads to FILE [assume single-end]\n"
-"  -t        copy RG, BC and QT tags to the %s header line\n",
+"  -s FILE              write singleton reads to FILE [assume single-end]\n"
+"  -t                   copy RG, BC and QT tags to the %s header line\n",
+    fq ? "FASTQ" : "FASTA");
+    fprintf(to,
+"  -T TAGLIST           copy arbitrary tags to the %s header line\n",
     fq ? "FASTQ" : "FASTA");
     if (fq) fprintf(to,
-"  -v INT    default quality score if not given in file [1]\n");
+"  -v INT               default quality score if not given in file [1]\n"
+"  -i                   add Illumina Casava 1.8 format entry to header (eg 1:N:0:ATCACG)\n"
+"  -c                   compression level [0..9] to use when creating gz or bgzf fastq files\n"
+"  --i1 FILE            write first index reads to FILE\n"
+"  --i2 FILE            write second index reads to FILE\n"
+"  --barcode-tag TAG    Barcode tag [default: " DEFAULT_BARCODE_TAG "]\n"
+"  --quality-tag TAG    Quality tag [default: " DEFAULT_QUALITY_TAG "]\n"
+"  --index-format STR   How to parse barcode and quality tags\n\n");
     sam_global_opt_help(to, "-.--.@");
+    fprintf(to,
+"   \n"
+"   The index-format string describes how to parse the barcode and quality tags, for example:\n"
+"   i14i8       the first 14 characters are index 1, the next 8 characters are index 2\n"
+"   n8i14       ignore the first 8 characters, and use the next 14 characters for index 1\n"
+"   If the tag contains a separator, then the numeric part can be replaced with '*' to mean\n"
+"   'read until the separator or end of tag', for example:\n"
+"   n*i*        ignore the left part of the tag until the separator, then use the second part\n"
+"               of the tag as index 1\n");
 }
 
 typedef enum { READ_UNKNOWN = 0, READ_1 = 1, READ_2 = 2 } readpart;
@@ -672,24 +709,110 @@ typedef struct bam2fq_opts {
     char *fnse;
     char *fnr[3];
     char *fn_input; // pointer to input filename in argv do not free
-    bool has12, use_oq, copy_tags;
-    int flag_on, flag_off;
+    bool has12, has12always, use_oq, copy_tags, illumina_tag;
+    int flag_on, flag_off, flag_alloff;
     sam_global_args ga;
     fastfile filetype;
     int def_qual;
+    char *barcode_tag;
+    char *quality_tag;
+    char *index_file[2];
+    char *index_format;
+    char *extra_tags;
+    char compression_level;
 } bam2fq_opts_t;
 
 typedef struct bam2fq_state {
     samFile *fp;
-    FILE *fpse;
-    FILE *fpr[3];
+    BGZF *fpse;
+    BGZF *fpr[3];
+    BGZF *fpi[2];
+    BGZF *hstdout;
     bam_hdr_t *h;
-    bool has12, use_oq, copy_tags;
-    int flag_on, flag_off;
+    bool has12, use_oq, copy_tags, illumina_tag;
+    int flag_on, flag_off, flag_alloff;
     fastfile filetype;
     int def_qual;
+    klist_t(ktaglist) *taglist;
+    char *index_sequence;
+    char compression_level;
 } bam2fq_state_t;
 
+/*
+ * Get and decode the read from a BAM record.
+ *
+ * TODO: htslib really needs an interface for this.  Consider this or perhaps
+ * bam_get_seq_str (current vs original orientation) and bam_get_qual_str
+ * functions as string formatted equivalents to bam_get_{seq,qual}?
+ */
+
+/*
+ * Reverse a string in place.
+ * From http://stackoverflow.com/questions/8534274/is-the-strrev-function-not-available-in-linux.
+ * Author Sumit-naik: http://stackoverflow.com/users/4590926/sumit-naik
+ */
+static char *reverse(char *str)
+{
+    int i = strlen(str)-1,j=0;
+    char ch;
+    while (i>j) {
+        ch = str[i];
+        str[i]= str[j];
+        str[j] = ch;
+        i--;
+        j++;
+    }
+    return str;
+}
+
+/* return the read, reverse complemented if necessary */
+static char *get_read(const bam1_t *rec)
+{
+    int len = rec->core.l_qseq + 1;
+    char *read = calloc(1, len);
+    char *seq = (char *)bam_get_seq(rec);
+    int n;
+
+    if (!read) return NULL;
+
+    for (n=0; n < rec->core.l_qseq; n++) {
+        if (rec->core.flag & BAM_FREVERSE) read[n] = seq_nt16_str[seq_comp_table[bam_seqi(seq,n)]];
+        else                               read[n] = seq_nt16_str[bam_seqi(seq,n)];
+    }
+    if (rec->core.flag & BAM_FREVERSE) reverse(read);
+    return read;
+}
+
+/*
+ * get and decode the quality from a BAM record
+ */
+static int get_quality(const bam1_t *rec, char **qual_out)
+{
+    char *quality = calloc(1, rec->core.l_qseq + 1);
+    char *q = (char *)bam_get_qual(rec);
+    int n;
+
+    if (!quality) return -1;
+
+    if (*q == '\xff') {
+        free(quality);
+        *qual_out = NULL;
+        return 0;
+    }
+
+    for (n=0; n < rec->core.l_qseq; n++) {
+        quality[n] = q[n]+33;
+    }
+    if (rec->core.flag & BAM_FREVERSE) reverse(quality);
+    *qual_out = quality;
+    return 0;
+}
+
+//
+// End of htslib complaints
+//
+
+
 static readpart which_readpart(const bam1_t *b)
 {
     if ((b->core.flag & BAM_FREAD1) && !(b->core.flag & BAM_FREAD2)) {
@@ -701,91 +824,284 @@ static readpart which_readpart(const bam1_t *b)
     }
 }
 
+/*
+ * parse the length part from the index-format string
+ */
+static int getLength(char **s)
+{
+    int n = 0;
+    while (**s) {
+        if (**s == '*') { n=-1; (*s)++; break; }
+        if ( !isdigit(**s)) break;
+        n = n*10 + ((**s)-'0');
+        (*s)++;
+    }
+    return n;
+}
+
+static bool copy_tag(const char *tag, const bam1_t *rec, kstring_t *linebuf)
+{
+    uint8_t *s = bam_aux_get(rec, tag);
+    if (s) {
+        char aux_type = *s;
+        switch (aux_type) {
+            case 'C':
+            case 'S': aux_type = 'I'; break;
+            case 'c':
+            case 's': aux_type = 'i'; break;
+            case 'd': aux_type = 'f'; break;
+        }
+
+        // Ensure space.  Need 6 chars + length of tag.  Max length of
+        // i is 16, A is 21, B currently 26, Z is unknown, so
+        // have to check that one later.
+        if (ks_resize(linebuf, ks_len(linebuf) + 64) < 0) return false;
+
+        kputc('\t', linebuf);
+        kputsn(tag, 2, linebuf);
+        kputc(':', linebuf);
+        kputc(aux_type=='I'? 'i': aux_type, linebuf);
+        kputc(':', linebuf);
+        switch (aux_type) {
+            case 'H':
+            case 'Z':
+                if (kputs(bam_aux2Z(s), linebuf) < 0) return false;
+                break;
+            case 'i': kputw(bam_aux2i(s), linebuf); break;
+            case 'I': kputuw(bam_aux2i(s), linebuf); break;
+            case 'A': kputc(bam_aux2A(s), linebuf); break;
+            case 'f': kputd(bam_aux2f(s), linebuf); break;
+            case 'B': kputs("*** Unhandled aux type ***", linebuf); return false;
+            default:  kputs("*** Unknown aux type ***", linebuf); return false;
+       }
+    }
+    return true;
+}
+
+static int insert_index_sequence_into_linebuf(char *index_sequence, kstring_t *linebuf, bam1_t *rec)
+{
+    if (!index_sequence) return 0;
+
+    kstring_t new = {0,0,NULL};
+    if (linebuf->s) {
+        char *s = strchr(linebuf->s, '\n');
+        if (s) {
+            if (ks_resize(&new, linebuf->l + strlen(index_sequence) + 16) < 0)
+                return -1;
+            *s = 0;
+            kputs(linebuf->s, &new);
+            kputc(' ', &new);
+            readpart readpart = which_readpart(rec);
+            if (readpart == READ_1) kputc('1', &new);
+            else if (readpart == READ_2) kputc('2', &new);
+            else kputc('0', &new);
+
+            kputc(':', &new);
+            if (rec->core.flag & BAM_FQCFAIL) kputc('Y', &new);
+            else                              kputc('N', &new);
+
+            kputs(":0:", &new);
+            kputs(index_sequence, &new);
+            kputc('\n', &new);
+            kputs(s+1, &new);
+            free(ks_release(linebuf));
+            linebuf->s = new.s; linebuf->l = new.l; linebuf->m = new.m;
+        }
+    }
+    return 0;
+}
+
+static bool make_fq_line(const bam1_t *rec, char *seq, char *qual, kstring_t *linebuf, const bam2fq_state_t *state)
+{
+    int i;
+
+    linebuf->l = 0;
+    // Write read name
+    if (kputc(state->filetype == FASTA? '>' : '@', linebuf) < 0) return false;
+    if (kputs(bam_get_qname(rec), linebuf) < 0) return false;
+    // Add the /1 /2 if requested
+    if (state->has12) {
+        readpart readpart = which_readpart(rec);
+        if (readpart == READ_1) {
+            if (kputs("/1", linebuf) < 0) return false;
+        } else if (readpart == READ_2) {
+            if (kputs("/2", linebuf) < 0) return false;
+        }
+    }
+    if (state->copy_tags) {
+        for (i = 0; copied_tags[i]; ++i) {
+            if (!copy_tag(copied_tags[i], rec, linebuf)) {
+                fprintf(stderr, "Problem copying aux tags: [%s]\n", linebuf->s);
+                return false;
+            }
+        }
+    }
+
+    if (state->taglist->size) {
+        kliter_t(ktaglist) *p;
+        for (p = kl_begin(state->taglist); p != kl_end(state->taglist); p = kl_next(p)) {
+            if (!copy_tag(kl_val(p), rec, linebuf)) {
+                fprintf(stderr, "Problem copying aux tags: [%s]\n", linebuf->s);
+                return false;
+            }
+        }
+    }
+
+    if (kputc('\n', linebuf) < 0) return false;
+    if (kputs(seq, linebuf) < 0) return false;
+    if (kputc('\n', linebuf) < 0) return false;
+
+    if (state->filetype == FASTQ) {
+        // Write quality
+        if (kputs("+\n", linebuf) < 0) return false;
+        if (qual && *qual) {
+            if (kputs(qual, linebuf) < 0) return false;
+        } else {
+            int len = strlen(seq);
+            if (ks_resize(linebuf, ks_len(linebuf) + len + 1) < 0) return false;
+            for (i = 0; i < len; ++i) {
+                kputc(33 + state->def_qual, linebuf);
+            }
+        }
+        if (kputc('\n', linebuf) < 0) return false;
+    }
+    return true;
+}
+
+/*
+ * Create FASTQ lines from the barcode tag using the index-format 
+ */
+static bool tags2fq(bam1_t *rec, bam2fq_state_t *state, const bam2fq_opts_t* opts)
+{
+    uint8_t *p;
+    char *ifmt = opts->index_format;
+    char *tag = NULL;
+    char *qual = NULL;
+    char *sub_tag = NULL;
+    char *sub_qual = NULL;
+    size_t tag_len;
+    int file_number = 0;
+    kstring_t linebuf = { 0, 0, NULL }; // Buffer
+
+
+    // read barcode tag
+    p = bam_aux_get(rec,opts->barcode_tag);
+    if (p) tag = bam_aux2Z(p);
+
+    if (!tag) return true; // there is no tag
+
+    tag_len = strlen(tag);
+    sub_tag = calloc(1, tag_len + 1);
+    if (!sub_tag) goto fail;
+    sub_qual = calloc(1, tag_len + 1);
+    if (!sub_qual) goto fail;
+
+    // read quality tag
+    p = bam_aux_get(rec, opts->quality_tag);
+    if (p) qual = bam_aux2Z(p);
+
+    // Parse the index-format string
+    while (*ifmt) {
+        if (file_number > 1) break;     // shouldn't happen if we've validated paramaters correctly
+        char action = *ifmt;        // should be 'i' or 'n'
+        ifmt++; // skip over action
+        int index_len = getLength(&ifmt);
+        int n = 0;
+
+        if (index_len < 0) {
+            // read until separator
+            while (isalpha(*tag)) {
+                sub_tag[n] = *tag++;
+                if (qual) sub_qual[n] = *qual++;
+                n++;
+            }
+            if (*tag) { // skip separator
+                tag++;
+                if (qual) qual++;
+            }
+        } else {
+            // read index_len characters
+            while (index_len-- && *tag) {
+                sub_tag[n] = *tag++;
+                if (qual) sub_qual[n] = *qual++;
+                n++;
+            }
+        }
+        sub_tag[n] = '\0';
+        sub_qual[n] = '\0';
+
+        if (action=='i' && *sub_tag && state->fpi[file_number]) {
+            //if (file_number==0) state->index_sequence = strdup(sub_tag);    // we're going to need this later...
+            state->index_sequence = strdup(sub_tag);    // we're going to need this later...
+            if (!state->index_sequence) goto fail;
+            if (!make_fq_line(rec, sub_tag, sub_qual, &linebuf, state)) goto fail;
+            if (state->illumina_tag) {
+                if (insert_index_sequence_into_linebuf(state->index_sequence, &linebuf, rec) < 0) {
+                    goto fail;
+                }
+            }
+            if (bgzf_write(state->fpi[file_number++], linebuf.s, linebuf.l) < 0)
+                goto fail;
+        }
+
+    }
+
+    free(sub_qual); free(sub_tag);
+    free(linebuf.s);
+    return true;
+
+ fail:
+    perror(__func__);
+    free(sub_qual); free(sub_tag);
+    free(linebuf.s);
+    return true;
+}
+
 // Transform a bam1_t record into a string with the FASTQ representation of it
 // @returns false for error, true for success
 static bool bam1_to_fq(const bam1_t *b, kstring_t *linebuf, const bam2fq_state_t *state)
 {
-    int i;
     int32_t qlen = b->core.l_qseq;
     assert(qlen >= 0);
-    uint8_t *seq;
-    uint8_t *qual = bam_get_qual(b);
     const uint8_t *oq = NULL;
+    char *qual = NULL;
+
+    char *seq = get_read(b);
+    if (!seq) return false;
+
     if (state->use_oq) {
         oq = bam_aux_get(b, "OQ");
-        if (oq) oq++; // skip tag type
-    }
-    bool has_qual = (qual[0] != 0xff || (state->use_oq && oq)); // test if there is quality
-
-    linebuf->l = 0;
-    // Write read name
-    readpart readpart = which_readpart(b);
-    kputc(state->filetype == FASTA? '>' : '@', linebuf);
-    kputs(bam_get_qname(b), linebuf);
-    // Add the /1 /2 if requested
-    if (state->has12) {
-        if (readpart == READ_1) kputs("/1", linebuf);
-        else if (readpart == READ_2) kputs("/2", linebuf);
-    }
-    if (state->copy_tags) {
-        for (i = 0; copied_tags[i]; ++i) {
-            uint8_t *s;
-            if ((s = bam_aux_get(b, copied_tags[i])) != 0) {
-                kputc('\t', linebuf);
-                kputsn(copied_tags[i], 2, linebuf);
-                kputsn(":Z:", 3, linebuf);
-                kputs(bam_aux2Z(s), linebuf);
+        if (oq) {
+            oq++; 
+            qual = strdup(bam_aux2Z(oq));
+            if (!qual) goto fail;
+            if (b->core.flag & BAM_FREVERSE) { // read is reverse complemented
+                reverse(qual);
             }
         }
-    }
-    kputc('\n', linebuf);
-
-    seq = bam_get_seq(b);
-
-    if (b->core.flag & BAM_FREVERSE) { // read is reverse complemented
-        for (i = qlen-1; i > -1; --i) {
-            char c = seq_nt16_str[seq_comp_table[bam_seqi(seq,i)]];
-            kputc(c, linebuf);
-        }
     } else {
-        for (i = 0; i < qlen; ++i) {
-            char c = seq_nt16_str[bam_seqi(seq,i)];
-            kputc(c, linebuf);
-        }
+        if (get_quality(b, &qual) < 0) goto fail;
     }
-    kputc('\n', linebuf);
 
-    if (state->filetype == FASTQ) {
-        // Write quality
-        kputs("+\n", linebuf);
-        if (has_qual) {
-            if (state->use_oq && oq) {
-                if (b->core.flag & BAM_FREVERSE) { // read is reverse complemented
-                    for (i = qlen-1; i > -1; --i) {
-                        kputc(oq[i], linebuf);
-                    }
-                } else {
-                    kputs((char*)oq, linebuf);
-                }
-            } else {
-                if (b->core.flag & BAM_FREVERSE) { // read is reverse complemented
-                    for (i = qlen-1; i > -1; --i) {
-                        kputc(33 + qual[i], linebuf);
-                    }
-                } else {
-                    for (i = 0; i < qlen; ++i) {
-                        kputc(33 + qual[i], linebuf);
-                    }
-                }
-            }
-        } else {
-            for (i = 0; i < qlen; ++i) {
-                kputc(33 + state->def_qual, linebuf);
-            }
-        }
-        kputc('\n', linebuf);
-    }
+    if (!make_fq_line(b, seq, qual, linebuf, state)) goto fail;
+
+    free(qual);
+    free(seq);
     return true;
+
+ fail:
+    free(seq);
+    free(qual);
+    return false;
+}
+
+static void free_opts(bam2fq_opts_t *opts)
+{
+    free(opts->barcode_tag);
+    free(opts->quality_tag);
+    free(opts->index_format);
+    free(opts->extra_tags);
+    free(opts);
 }
 
 // return true if valid
@@ -794,43 +1110,109 @@ static bool parse_opts(int argc, char *argv[], bam2fq_opts_t** opts_out)
     // Parse args
     bam2fq_opts_t* opts = calloc(1, sizeof(bam2fq_opts_t));
     opts->has12 = true;
+    opts->has12always = false;
     opts->filetype = FASTQ;
     opts->def_qual = 1;
+    opts->barcode_tag = NULL;
+    opts->quality_tag = NULL;
+    opts->index_format = NULL;
+    opts->index_file[0] = NULL;
+    opts->index_file[1] = NULL;
+    opts->extra_tags = NULL;
+    opts->compression_level = 1;
 
     int c;
     sam_global_args_init(&opts->ga);
     static const struct option lopts[] = {
         SAM_OPT_GLOBAL_OPTIONS('-', 0, '-', '-', 0, '@'),
+        {"i1", required_argument, NULL, 1},
+        {"I1", required_argument, NULL, 1},
+        {"i2", required_argument, NULL, 2},
+        {"I2", required_argument, NULL, 2},
+        {"if", required_argument, NULL, 3},
+        {"IF", required_argument, NULL, 3},
+        {"index-format", required_argument, NULL, 3},
+        {"barcode-tag", required_argument, NULL, 'b'},
+        {"quality-tag", required_argument, NULL, 'q'},
         { NULL, 0, NULL, 0 }
     };
-    while ((c = getopt_long(argc, argv, "0:1:2:f:F:nOs:tv:@:", lopts, NULL)) > 0) {
+    while ((c = getopt_long(argc, argv, "0:1:2:f:F:G:niNOs:c:tT:v:@:", lopts, NULL)) > 0) {
         switch (c) {
+            case 'b': opts->barcode_tag = strdup(optarg); break;
+            case 'q': opts->quality_tag = strdup(optarg); break;
+            case  1 : opts->index_file[0] = optarg; break;
+            case  2 : opts->index_file[1] = optarg; break;
+            case  3 : opts->index_format = strdup(optarg); break;
             case '0': opts->fnr[0] = optarg; break;
             case '1': opts->fnr[1] = optarg; break;
             case '2': opts->fnr[2] = optarg; break;
             case 'f': opts->flag_on |= strtol(optarg, 0, 0); break;
             case 'F': opts->flag_off |= strtol(optarg, 0, 0); break;
+            case 'G': opts->flag_alloff |= strtol(optarg, 0, 0); break;
             case 'n': opts->has12 = false; break;
+            case 'N': opts->has12always = true; break;
             case 'O': opts->use_oq = true; break;
             case 's': opts->fnse = optarg; break;
             case 't': opts->copy_tags = true; break;
+            case 'i': opts->illumina_tag = true; break;
+            case 'c': opts->compression_level = atoi(optarg); break;
+            case 'T': opts->extra_tags = strdup(optarg); break;
             case 'v': opts->def_qual = atoi(optarg); break;
-            case '?': bam2fq_usage(stderr, argv[0]); free(opts); return false;
+            case '?': bam2fq_usage(stderr, argv[0]); free_opts(opts); return false;
             default:
                 if (parse_sam_global_opt(c, optarg, lopts, &opts->ga) != 0) {
-                    bam2fq_usage(stderr, argv[0]); free(opts); return false;
+                    bam2fq_usage(stderr, argv[0]); free_opts(opts); return false;
                 }
                 break;
         }
     }
 
     if (opts->fnr[1] || opts->fnr[2]) opts->has12 = false;
+    if (opts->has12always) opts->has12 = true;
+
+    if (!opts->barcode_tag) opts->barcode_tag = strdup(DEFAULT_BARCODE_TAG);
+    if (!opts->quality_tag) opts->quality_tag = strdup(DEFAULT_QUALITY_TAG);
+
+    int nIndex = 0;
+    if (opts->index_format) {
+        char *s;
+        for (s = opts->index_format; *s; s++) {
+            if (*s == 'i') nIndex++;
+        }
+    }
+    if (nIndex>2) {
+        fprintf(stderr,"Invalid index format: more than 2 indexes\n");
+        bam2fq_usage(stderr, argv[0]);
+        free_opts(opts);
+        return false;
+    }
+
+    if (opts->index_file[1] && !opts->index_file[0]) {
+        fprintf(stderr, "Index one specified, but index two not given\n");
+        bam2fq_usage(stderr, argv[0]);
+        free_opts(opts);
+        return false;
+    }
+
+    if (nIndex==2 && !opts->index_file[1]) {
+        fprintf(stderr, "index_format specifies two indexes, but only one index file given\n");
+        bam2fq_usage(stderr, argv[0]);
+        free_opts(opts);
+        return false;
+    }
+
+    if (nIndex==1 && !opts->index_file[0]) {
+        fprintf(stderr, "index_format specifies an index, but no index file given\n");
+        bam2fq_usage(stderr, argv[0]);
+        free_opts(opts);
+        return false;
+    }
 
     if (opts->def_qual < 0 || 93 < opts->def_qual) {
         fprintf(stderr, "Invalid -v default quality %i, allowed range 0 to 93\n", opts->def_qual);
         bam2fq_usage(stderr, argv[0]);
-        free(opts);
-        return true;
+        free_opts(opts);
+        return false;
     }
 
     const char* type_str = argv[0];
@@ -841,20 +1223,21 @@ static bool parse_opts(int argc, char *argv[], bam2fq_opts_t** opts_out)
     } else {
         print_error("bam2fq", "Unrecognised type call \"%s\", this should be impossible... but you managed it!", type_str);
         bam2fq_usage(stderr, argv[0]);
-        free(opts);
+        free_opts(opts);
         return false;
     }
 
     if ((argc - (optind)) == 0) {
+        fprintf(stderr, "No input file specified.\n");
         bam2fq_usage(stdout, argv[0]);
-        free(opts);
+        free_opts(opts);
         return false;
     }
 
     if ((argc - (optind)) != 1) {
         fprintf(stderr, "Too many arguments.\n");
         bam2fq_usage(stderr, argv[0]);
-        free(opts);
+        free_opts(opts);
         return false;
     }
     opts->fn_input = argv[optind];
@@ -862,16 +1245,55 @@ static bool parse_opts(int argc, char *argv[], bam2fq_opts_t** opts_out)
     return true;
 }
 
+static BGZF *open_fqfile(char *filename, int c)
+{
+    char mode[4] = "w";
+    size_t len = strlen(filename);
+
+    mode[2] = 0; mode[3] = 0;
+    if (len > 3 && strstr(filename + (len - 3),".gz")) {
+        mode[1] = 'g'; mode[2] = c+'0';
+    } else if ((len > 4 && strstr(filename + (len - 4),".bgz"))
+               || (len > 5 && strstr(filename + (len - 5),".bgzf"))) {
+        mode[1] = c+'0';
+    } else {
+        mode[1] = 'u';
+    }
+
+    return bgzf_open(filename,mode);
+}
+
 static bool init_state(const bam2fq_opts_t* opts, bam2fq_state_t** state_out)
 {
     bam2fq_state_t* state = calloc(1, sizeof(bam2fq_state_t));
     state->flag_on = opts->flag_on;
     state->flag_off = opts->flag_off;
+    state->flag_alloff = opts->flag_alloff;
     state->has12 = opts->has12;
     state->use_oq = opts->use_oq;
+    state->illumina_tag = opts->illumina_tag;
     state->copy_tags = opts->copy_tags;
     state->filetype = opts->filetype;
     state->def_qual = opts->def_qual;
+    state->index_sequence = NULL;
+    state->hstdout = bgzf_dopen(fileno(stdout), "wu");
+    state->compression_level = opts->compression_level;
+
+    state->taglist = kl_init(ktaglist);
+    if (opts->extra_tags) {
+        char *save_p;
+        char *s = strtok_r(opts->extra_tags, ",", &save_p);
+        while (s) {
+            if (strlen(s) != 2) {
+                fprintf(stderr, "Parsing extra tags - '%s' is not two characters\n", s);
+                free(state);
+                return false;
+            }
+            char **et = kl_pushp(ktaglist, state->taglist);
+            *et = s;
+            s = strtok_r(NULL, ",", &save_p);
+        }
+    }
 
     state->fp = sam_open(opts->fn_input, "r");
     if (state->fp == NULL) {
@@ -882,7 +1304,7 @@ static bool init_state(const bam2fq_opts_t* opts, bam2fq_state_t** state_out)
     if (opts->ga.nthreads > 0)
         hts_set_threads(state->fp, opts->ga.nthreads);
     uint32_t rf = SAM_QNAME | SAM_FLAG | SAM_SEQ | SAM_QUAL;
-    if (opts->use_oq) rf |= SAM_AUX;
+    if (opts->use_oq || opts->extra_tags || opts->index_file[0]) rf |= SAM_AUX;
     if (hts_set_opt(state->fp, CRAM_OPT_REQUIRED_FIELDS, rf)) {
         fprintf(stderr, "Failed to set CRAM_OPT_REQUIRED_FIELDS value\n");
         free(state);
@@ -894,7 +1316,7 @@ static bool init_state(const bam2fq_opts_t* opts, bam2fq_state_t** state_out)
         return false;
     }
     if (opts->fnse) {
-        state->fpse = fopen(opts->fnse,"w");
+        state->fpse = open_fqfile(opts->fnse, state->compression_level);
         if (state->fpse == NULL) {
             print_error_errno("bam2fq", "Cannot write to singleton file \"%s\"", opts->fnse);
             free(state);
@@ -905,14 +1327,25 @@ static bool init_state(const bam2fq_opts_t* opts, bam2fq_state_t** state_out)
     int i;
     for (i = 0; i < 3; ++i) {
         if (opts->fnr[i]) {
-            state->fpr[i] = fopen(opts->fnr[i], "w");
+            state->fpr[i] = open_fqfile(opts->fnr[i], state->compression_level);
             if (state->fpr[i] == NULL) {
                 print_error_errno("bam2fq", "Cannot write to r%d file \"%s\"", i, opts->fnr[i]);
                 free(state);
                 return false;
             }
         } else {
-            state->fpr[i] = stdout;
+            state->fpr[i] = state->hstdout;
+        }
+    }
+    for (i = 0; i < 2; i++) {
+        state->fpi[i] = NULL;
+        if (opts->index_file[i]) {
+            state->fpi[i] = open_fqfile(opts->index_file[i], state->compression_level);
+            if (state->fpi[i] == NULL) {
+                print_error_errno("bam2fq", "Cannot write to i%d file \"%s\"", i+1, opts->index_file[i]);
+                free(state);
+                return false;
+            }
         }
     }
 
@@ -932,11 +1365,23 @@ static bool destroy_state(const bam2fq_opts_t *opts, bam2fq_state_t *state, int*
     bool valid = true;
     bam_hdr_destroy(state->h);
     check_sam_close("bam2fq", state->fp, opts->fn_input, "file", status);
-    if (state->fpse && fclose(state->fpse)) { print_error_errno("bam2fq", "Error closing singleton file \"%s\"", opts->fnse); valid = false; }
+    if (state->fpse && bgzf_close(state->fpse)) { print_error_errno("bam2fq", "Error closing singleton file \"%s\"", opts->fnse); valid = false; }
     int i;
     for (i = 0; i < 3; ++i) {
-        if (state->fpr[i] != stdout && fclose(state->fpr[i])) { print_error_errno("bam2fq", "Error closing r%d file \"%s\"", i, opts->fnr[i]); valid = false; }
+        if (state->fpr[i] == state->hstdout) {
+            if (i==0 && bgzf_close(state->fpr[i])) { print_error_errno("bam2fq", "Error closing STDOUT"); valid = false; }
+        } else {
+            if (bgzf_close(state->fpr[i])) { print_error_errno("bam2fq", "Error closing r%d file \"%s\"", i, opts->fnr[i]); valid = false; }
+        }
     }
+    for (i = 0; i < 2; i++) {
+        if (state->fpi[i] && bgzf_close(state->fpi[i])) { 
+            print_error_errno("bam2fq", "Error closing i%d file \"%s\"", i+1, opts->index_file[i]);
+            valid = false;
+        }
+    }
+    kl_destroy(ktaglist,state->taglist);
+    free(state->index_sequence);
     free(state);
     return valid;
 }
@@ -945,12 +1390,15 @@ static inline bool filter_it_out(const bam1_t *b, const bam2fq_state_t *state)
 {
     return (b->core.flag&(BAM_FSECONDARY|BAM_FSUPPLEMENTARY) // skip secondary and supplementary alignments
         || (b->core.flag&(state->flag_on)) != state->flag_on // or reads indicated by filter flags
-        || (b->core.flag&(state->flag_off)) != 0);
+        || (b->core.flag&(state->flag_off)) != 0
+        || (b->core.flag&(state->flag_alloff) && (b->core.flag&(state->flag_alloff)) == state->flag_alloff));
 
 }
 
-static bool bam2fq_mainloop_singletontrack(bam2fq_state_t *state)
+static bool bam2fq_mainloop(bam2fq_state_t *state, bam2fq_opts_t* opts)
 {
+    int n;
+    bam1_t *records[3];
     bam1_t* b = bam_init1();
     char *current_qname = NULL;
     int64_t n_reads = 0, n_singletons = 0; // Statistics
@@ -958,35 +1406,47 @@ static bool bam2fq_mainloop_singletontrack(bam2fq_state_t *state)
     int score[3];
     int at_eof;
     if (b == NULL ) {
-        perror("[bam2fq_mainloop_singletontrack] Malloc error for bam record buffer.");
+        perror("[bam2fq_mainloop] Malloc error for bam record buffer.");
         return false;
     }
 
     bool valid = true;
     while (true) {
-        at_eof = sam_read1(state->fp, state->h, b) < 0;
+        int res = sam_read1(state->fp, state->h, b);
+        if (res < -1) {
+            fprintf(stderr, "[bam2fq_mainloop] Failed to read bam record.\n");
+            return false;
+        }
+        at_eof = res < 0;
 
         if (!at_eof && filter_it_out(b, state)) continue;
         if (!at_eof) ++n_reads;
 
         if (at_eof || !current_qname || (strcmp(current_qname, bam_get_qname(b)) != 0)) {
             if (current_qname) {
+                if (state->illumina_tag) {
+                    for (n=0; valid && n<3; n++) {
+                        if (insert_index_sequence_into_linebuf(state->index_sequence, &linebuf[n], records[n]) < 0) valid = false;
+                    }
+                    if (!valid) break;
+                }
+                free(state->index_sequence); state->index_sequence = NULL;
                 if (score[1] > 0 && score[2] > 0) {
                     // print linebuf[1] to fpr[1], linebuf[2] to fpr[2]
-                    if (fputs(linebuf[1].s, state->fpr[1]) == EOF) { valid = false; break; }
-                    if (fputs(linebuf[2].s, state->fpr[2]) == EOF) { valid = false; break; }
-                } else if (score[1] > 0 || score[2] > 0) {
+                    if (bgzf_write(state->fpr[1], linebuf[1].s, linebuf[1].l) < 0) { valid = false; break; }
+                    if (bgzf_write(state->fpr[2], linebuf[2].s, linebuf[2].l) < 0) { valid = false; break; }
+                } else if ((score[1] > 0 || score[2] > 0) && state->fpse) {
                     // print whichever one exists to fpse
                     if (score[1] > 0) {
-                        if (fputs(linebuf[1].s, state->fpse) == EOF) { valid = false; break; }
+                        if (bgzf_write(state->fpse, linebuf[1].s, linebuf[1].l) < 0) { valid = false; break; }
                     } else {
-                        if (fputs(linebuf[2].s, state->fpse) == EOF) { valid = false; break; }
+                        if (bgzf_write(state->fpse, linebuf[2].s, linebuf[2].l) < 0) { valid = false; break; }
                     }
                     ++n_singletons;
                 }
                 if (score[0]) { // TODO: check this
                     // print linebuf[0] to fpr[0]
-                    if (fputs(linebuf[0].s, state->fpr[0]) == EOF) { valid = false; break; }
+                    if (bgzf_write(state->fpr[0], linebuf[0].s, linebuf[0].l) < 0) { valid = false; break; }
                 }
             }
 
@@ -994,12 +1454,15 @@ static bool bam2fq_mainloop_singletontrack(bam2fq_state_t *state)
 
             free(current_qname);
             current_qname = strdup(bam_get_qname(b));
+            if (!current_qname) { valid = false; break; }
             score[0] = score[1] = score[2] = 0;
         }
 
         // Prefer a copy of the read that has base qualities
         int b_score = bam_get_qual(b)[0] != 0xff? 2 : 1;
         if (b_score > score[which_readpart(b)]) {
+            if (state->fpi[0]) if (!tags2fq(b, state, opts)) return false;
+            records[which_readpart(b)] = b;
             if(!bam1_to_fq(b, &linebuf[which_readpart(b)], state)) {
                 fprintf(stderr, "[%s] Error converting read to FASTA/Q\n", __func__);
                 return false;
@@ -1009,7 +1472,7 @@ static bool bam2fq_mainloop_singletontrack(bam2fq_state_t *state)
     }
     if (!valid)
     {
-        perror("[bam2fq_mainloop_singletontrack] Error writing to FASTx files.");
+        perror("[bam2fq_mainloop] Error writing to FASTx files.");
     }
     bam_destroy1(b);
     free(current_qname);
@@ -1022,30 +1485,6 @@ static bool bam2fq_mainloop_singletontrack(bam2fq_state_t *state)
     return valid;
 }
 
-static bool bam2fq_mainloop(bam2fq_state_t *state)
-{
-    // process a name collated BAM into fastq
-    bam1_t* b = bam_init1();
-    if (b == NULL) {
-        perror(NULL);
-        return false;
-    }
-    int64_t n_reads = 0; // Statistics
-    kstring_t linebuf = { 0, 0, NULL }; // Buffer
-    while (sam_read1(state->fp, state->h, b) >= 0) {
-        if (filter_it_out(b, state)) continue;
-        ++n_reads;
-
-        if (!bam1_to_fq(b, &linebuf, state)) return false;
-        fputs(linebuf.s, state->fpr[which_readpart(b)]);
-    }
-    free(linebuf.s);
-    bam_destroy1(b);
-
-    fprintf(stderr, "[M::%s] processed %" PRId64 " reads\n", __func__, n_reads);
-    return true;
-}
-
 int main_bam2fq(int argc, char *argv[])
 {
     int status = EXIT_SUCCESS;
@@ -1057,15 +1496,11 @@ int main_bam2fq(int argc, char *argv[])
 
     if (!init_state(opts, &state)) return EXIT_FAILURE;
 
-    if (state->fpse) {
-        if (!bam2fq_mainloop_singletontrack(state)) status = EXIT_FAILURE;
-    } else {
-        if (!bam2fq_mainloop(state)) status = EXIT_FAILURE;
-    }
+    if (!bam2fq_mainloop(state,opts)) status = EXIT_FAILURE;
 
     if (!destroy_state(opts, state, &status)) return EXIT_FAILURE;
     sam_global_args_free(&opts->ga);
-    free(opts);
+    free_opts(opts);
 
     return status;
 }
